<?php

declare(strict_types=1);

namespace Drupal\Tests\joinup_stats\Kernel;

use Drupal\KernelTests\KernelTestBase;
use Drupal\Tests\joinup_test\Traits\ConfigTestTrait;
use Drupal\Tests\sparql_entity_storage\Traits\SparqlConnectionTrait;
use Drupal\field\Entity\FieldConfig;
use Drupal\field\Entity\FieldStorageConfig;
use Drupal\file\Entity\File;

/**
 * Tests the download & visit count refresh.
 *
 * @group joinup_stats
 */
class RefreshCountersTest extends KernelTestBase {

  use ConfigTestTrait;
  use SparqlConnectionTrait;

  /**
   * Testing entities.
   *
   * @var array
   */
  protected $entities = [];

  /**
   * {@inheritdoc}
   */
  protected static $modules = [
    'cached_computed_field',
    'dynamic_entity_reference',
    'field',
    'file',
    'file_url',
    'joinup_stats',
    'joinup_stats_test',
    'matomo_reporting_api',
    'meta_entity',
    'node',
    'og',
    'options',
    'rdf_entity',
    'sparql_entity_storage',
    'system',
    'user',
  ];

  /**
   * {@inheritdoc}
   */
  protected function bootEnvironment() {
    parent::bootEnvironment();
    $this->setUpSparql();
  }

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    $this->installEntitySchema('meta_entity');
    $this->installEntitySchema('user');
    $this->installEntitySchema('node');
    $this->installEntitySchema('file');
    $this->installSchema('file', ['file_usage']);

    $this->installConfig([
      'cached_computed_field',
      'joinup_stats_test',
      'sparql_entity_storage',
    ]);
    FieldStorageConfig::create($this->getConfigData('field.storage.meta_entity.count'))->save();

    // Import the DownloadCount and VisitCount meta entity bundles from our
    // exported config.
    $this->importConfigs([
      'meta_entity.type.download_count',
      'meta_entity.type.visit_count',
    ]);
    $this->container->get('entity_type.bundle.info')->clearCachedBundles();

    FieldConfig::create($this->getConfigData('field.field.meta_entity.download_count.count'))
      ->setSetting('cache-max-age', 0)
      ->save();
    FieldConfig::create($this->getConfigData('field.field.meta_entity.visit_count.count'))
      ->setSetting('cache-max-age', 0)
      ->save();

    $this->createContent();
  }

  /**
   * Tests download & visit count refresh.
   */
  public function testRefreshCounters(): void {
    $queue = $this->container->get('cached_computed_field.manager')->getQueue();

    // Check that creating entities the meta entities are also created. Only
    // test one RDF entity and one node.
    $this->reloadEntities();
    $distro = $this->entities['rdf_entity']['http://example.com/distro/2'];
    $this->assertSame('http://example.com/distro/2', $distro->download_count->entity->target->target_id);
    $node = $this->entities['node'][3];
    $this->assertEquals(3, $node->visit_count->entity->target->target_id);

    // Check that before first cron run the queue is empty.
    $this->assertEquals(0, $queue->numberOfItems());

    // Run cron for the first time.
    cached_computed_field_cron();

    // Check that after first cron run all entities were queued.
    $this->assertEquals(7, $queue->numberOfItems());

    // Second cron run.
    cached_computed_field_cron();

    // Check that after second cron the queue has been consumed.
    $this->assertEquals(0, $queue->numberOfItems());

    // Check that the cached computed fields were updated.
    // @see \Drupal\joinup_stats_test\Mocks\TestQuery::getMockedResponseArray()
    $this->assertEquals(55, $this->entities['rdf_entity']['http://example.com/distro/1']->getDownloadCount());
    $this->assertEquals(2034, $this->entities['rdf_entity']['http://example.com/distro/2']->getDownloadCount());
    $this->assertEquals(0, $this->entities['rdf_entity']['http://example.com/distro/3']->getDownloadCount());
    $this->assertEquals(3846545, $this->entities['node'][1]->getVisitCount());
    $this->assertEquals(234, $this->entities['node'][2]->getVisitCount());
    $this->assertEquals(8766, $this->entities['node'][3]->getVisitCount());
    $this->assertEquals(334, $this->entities['node'][4]->getVisitCount());

    // Pretend that the stats were incremented on Matomo side.
    $this->container->get('state')->set('joinup_stats_test.increment', [
      'http://example.com/distro/1' => 1000,
      'http://example.com/distro/2' => 10000,
      'http://example.com/distro/3' => 39,
      '1' => 1000000,
      '2' => 10000,
      '3' => 20000,
      '4' => 200000,
    ]);
    sleep(1);

    // On third cron run the queue should be repopulated.
    cached_computed_field_cron();
    $this->assertEquals(7, $queue->numberOfItems());

    // Run again the cron and check that values were incremented. Entities need
    // reload in order to get the new values.
    cached_computed_field_cron();
    $this->reloadEntities();
    $this->assertEquals(1055, $this->entities['rdf_entity']['http://example.com/distro/1']->getDownloadCount());
    $this->assertEquals(12034, $this->entities['rdf_entity']['http://example.com/distro/2']->getDownloadCount());
    $this->assertEquals(39, $this->entities['rdf_entity']['http://example.com/distro/3']->getDownloadCount());
    $this->assertEquals(4846545, $this->entities['node'][1]->getVisitCount());
    $this->assertEquals(10234, $this->entities['node'][2]->getVisitCount());
    $this->assertEquals(28766, $this->entities['node'][3]->getVisitCount());
    $this->assertEquals(200334, $this->entities['node'][4]->getVisitCount());
  }

  /**
   * Refreshes the test entities.
   *
   * This ensures that we are getting the actual data from the meta entities and
   * not stale data from the entity storage cache.
   */
  protected function reloadEntities() {
    /** @var \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager */
    $entity_type_manager = $this->container->get('entity_type.manager');
    foreach ($this->entities as $entity_type_id => $entities) {
      foreach ($entities as $entity) {
        $this->entities[$entity_type_id][$entity->id()] = $entity_type_manager->getStorage($entity_type_id)->loadUnchanged($entity->id());
      }
    }
  }

  /**
   * Creates the test content.
   */
  protected function createContent(): void {
    $entity_type_manager = $this->container->get('entity_type.manager');
    foreach ($this->getContentData() as $entity_type_id => $items) {
      $storage = $entity_type_manager->getStorage($entity_type_id);
      foreach ($items as $values) {
        $entity = $storage->create($values);
        $entity->save();
        $this->entities[$entity_type_id][$entity->id()] = $entity;
      }
    }
  }

  /**
   * Returns test data for testing entities.
   *
   * @return array
   *   Test data.
   */
  protected function getContentData(): array {
    file_put_contents('public://d1', $this->randomString());
    file_put_contents('public://d3', $this->randomString());
    file_put_contents('public://d3', $this->randomString());

    return [
      'rdf_entity' => [
        [
          'id' => 'http://example.com/solution/1',
          'rid' => 'solution',
          'label' => 'Solution 1',
        ],
      ],
      // Enforce predictable NIDs.
      'node' => [
        [
          'type' => 'discussion',
          'nid' => 1,
          'title' => 'Discussion',
        ],
        [
          'type' => 'document',
          'nid' => 2,
          'title' => 'Document',
        ],
        [
          'type' => 'event',
          'nid' => 3,
          'title' => 'Event',
        ],
        [
          'type' => 'news',
          'nid' => 4,
          'title' => 'News',
        ],
      ],
      [
        'nid' => 5,
        'uri' => 'http://example.com/distro/1',
        'type' => 'distribution',
        'label' => 'Distro 1',
        'field_access_url' => File::create([
          'uri' => 'public://d1',
        ]),
        'og_audience' => 'http://example.com/solution/1',
        'field_state' => 'published',
      ],
      [
        'nid' => 6,
        'uri' => 'http://example.com/distro/2',
        'type' => 'distribution',
        'label' => 'Distro 2',
        'field_access_url' => File::create([
          'uri' => 'public://d2',
        ]),
        'field_state' => 'published',
      ],
      [
        'nid' => 7,
        'uri' => 'http://example.com/distro/3',
        'rid' => 'distribution',
        'label' => 'Distro 3',
        'field_access_url' => File::create([
          'uri' => 'public://d3',
        ]),
        'field_state' => 'published',
      ],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function tearDown(): void {
    $storage = $this->container->get('entity_type.manager')->getStorage('rdf_entity');
    $storage->delete($this->entities['rdf_entity']);
    parent::tearDown();
  }

}
